# -*- coding: utf-8 -*-
##############################################################################
#
#    Copyright (C) 2017-TODAY Eynes (<http://www.e-mips.com.ar>)
#
#    This is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from openerp import _, api, models, fields, exceptions
from openerp.addons import decimal_precision as dp
import logging
from lxml import etree

logger = logging.getLogger(__name__)


class stock_production_lot(models.Model):
    _inherit = 'stock.production.lot'

    @api.multi
    def action_update_last_location(self):
        for record in self:
            if record.quant_ids.ids:
                last_quant_id = max(record.quant_ids.ids)
                last_quant_data = self.env['stock.quant'].browse(last_quant_id)
                record.last_location_id = last_quant_data.location_id.id
            else:
                record.last_location_id = False

    @api.model
    def search_read(self, domain=None, fields=None, offset=0, limit=None, order=None):
        context = self._context
        cr = self._cr
        if context.get('pos'):
            # Modify domain
            prod_id = context.get('product_id')
            if self.env.user.is_contactcenter:  # Is contact, display all lots
                cr.execute("""
                    SELECT spl.id lot_id FROM stock_production_lot spl
                    JOIN stock_quant sq ON sq.lot_id = spl.id
                    JOIN stock_location sl ON sl.id = sq.location_id
                    LEFT JOIN stock_move sm ON sq.reservation_id = sm.id
                    WHERE sq.product_id=%s AND sl.usage ~ 'internal'
                    AND (sq.reservation_id IS NULL)  -- Not reserved or Incoming to our locations
                    GROUP BY spl.id HAVING sum(sq.qty)>0;
                """, (prod_id, ))
                lot_ids = [i[0] for i in cr.fetchall()]
                domain = [("id", "in", lot_ids)]
            else:
                loc_ids = tuple(context.get('loc_ids'))
                cr.execute("""
                SELECT spl.id lot_id FROM stock_production_lot spl
                JOIN stock_quant sq ON sq.lot_id = spl.id
                JOIN stock_location sl ON sl.id = sq.location_id
                LEFT JOIN stock_location sl_spl on sl_spl.id=spl.last_location_id
                LEFT JOIN stock_move sm ON sq.reservation_id = sm.id
                WHERE sq.product_id=%s AND sq.location_id IN %s AND sl.usage ~ 'internal'
                AND (sq.reservation_id IS NULL OR sm.location_dest_id IN %s)  -- Not reserved or Incoming to our locations
                AND sl_spl.usage ~ 'internal'
                GROUP BY spl.id HAVING sum(sq.qty)>0;
                """, (prod_id, loc_ids, loc_ids))
                lot_ids = [i[0] for i in cr.fetchall()]
                domain = [("id", "in", lot_ids)]
        res = super(stock_production_lot, self).search_read(domain, fields, offset, limit, order)
        return res

    @api.multi
    def check_stock_lot(self):
        """
        Ensure we have enough qty's available in the lot, if not return 0
        """
        self.ensure_one()
        location_ids = tuple([i for i in self._context.get('loc_ids') if i])
        product_id = self._context.get('product_id')
        self._cr.execute("""
        SELECT COALESCE(SUM(sq.qty), 0) qty FROM stock_production_lot spl
        JOIN stock_quant sq ON sq.lot_id = spl.id
        JOIN stock_location sl ON sl.id = sq.location_id
        WHERE sq.product_id=%s AND sl.usage ~ 'internal' AND spl.id=%s;
        """, (product_id, self.id))
        qty = self._cr.fetchone()[0]
        logger.info("Qty Available %s [product #%s, locations #%s, lot #%s]" % (qty, product_id, location_ids, self.id))
        return 0 if qty <= 0 else qty


class stock_warehouse(models.Model):
    _inherit = 'stock.warehouse'

    @api.model
    def create(self, vals):
        """
        Create a transit location inside the warehouse
        """
        route_id = self.env.ref('financing.route_transit_to_stock')
        vals.update({
            'route_ids': [(4, route_id.id, False)]
        })
        res = super(stock_warehouse, self).create(vals)
        location_obj = self.env['stock.location']
        sub_locations = [
                {'name': _('Transit'), 'active': True, 'is_transit': True},
        ]
        wh_loc_id = self.env['stock.location'].search([
            ('name','=',_(vals.get('code'))),
            ('location_id','=',self.env.ref('stock.stock_location_locations').id)
        ]).id
        for values in sub_locations:
            loc_vals = {
                'name': values['name'],
                'usage': 'internal',
                'location_id': wh_loc_id,
                'active': values['active'],
                'is_transit': values['is_transit'],
            }
            if vals.get('company_id'):
                loc_vals['company_id'] = vals.get('company_id')
            transit_loc_id = location_obj.create(loc_vals)
        # Add the push rule that when incoming to transit generate a new picking to stock manual op
        push_obj = self.env['stock.location.path']
        push_rules_list = []
        warehouse = res
        from_loc = transit_loc_id
        dest_loc = self.env['stock.location'].search([
            ('name', 'ilike', 'Stock'),
            ('location_id', '=', wh_loc_id)
        ])
        pick_type_id = self.env['stock.picking.type'].search([
            ('warehouse_id', '=', warehouse.id),
            ('code', '=', 'incoming')
        ])
        push_rules_list.append({
            'name': self._format_rulename(warehouse, from_loc, dest_loc),
            'location_from_id': from_loc.id,
            'location_dest_id': dest_loc.id,
            'route_id': route_id.id,
            'auto': 'manual',
            'picking_type_id': pick_type_id.id,
            'active': True,
            'warehouse_id': warehouse.id,
        })
        for push_rule in push_rules_list:
            push_obj.create(vals=push_rule)
        return res


class stock_move(models.Model):
    _inherit = 'stock.move'

    @api.model
    def default_get(self, fields):
        res = super(stock_move, self).default_get(fields)
        dlda = self._context.get('default_location_dest_id_aux')
        if dlda:
            res['location_dest_id'] = dlda
        return res

    @api.model
    def check_unicity_qty_available(self, operation_or_move):
        """
        Check quantity on lot to verify that has qty = 1
        if 'lot unique' is ok on product
        """
        lot = operation_or_move.lot_id
        if operation_or_move.product_id.lot_unique_ok and lot:
            qty = sum([x.qty + x.propagated_from_id.qty
                       for x in operation_or_move.lot_id.quant_ids])
            if (qty != 1 and qty != 0) or operation_or_move.product_qty != 1:
                raise exceptions.ValidationError(_(
                    "Product '%s' has active "
                    "'unique lot'\n"
                    "but with this move "
                    "you will have a quantity of "
                    "'%s' in lot '%s'"
                ) % (operation_or_move.product_id.name, qty, lot.name))
        return True

    @api.model
    def fields_view_get(self, view_id=None, view_type='form', toolbar=False, submenu=False):
        res = super(stock_move, self).fields_view_get(view_id=view_id, view_type=view_type, toolbar=toolbar, submenu=submenu)
        spt_id = self._context.get('default_picking_type_id')
        if spt_id:
            spt = self.env['stock.picking.type'].browse(spt_id)
            poscfg = self.env.user.pos_config
            if view_type == 'form' and poscfg and spt.code == 'internal':
                # On internal moves for pos users move origin is always from their stock location, dest is all available transit locations
                doc = etree.XML(res['arch'])
                suc_transit_ids = self.env['stock.location'].search([
                    ('usage', '=', 'internal'),
                    ('id', '!=', poscfg.stock_location_transit_id.id),
                    ('is_transit', '=', True),
                ]).ids
                for node in doc.xpath("//field[@name='location_id']"):
                    domain = "[('id', '=', %s)]" % poscfg.stock_location_id.id
                    node.set('domain', domain)
                for node in doc.xpath("//field[@name='location_dest_id']"):
                    # the 'Log a note' button is employee only
                    domain = "[('id', 'in', %s)]" % suc_transit_ids
                    node.set('domain', domain)
                res['arch'] = etree.tostring(doc)

        return res


class stock_location(models.Model):
    _inherit = 'stock.location'

    is_transit = fields.Boolean('Is transitory location?')
    is_contact = fields.Boolean('Is Contact Center?', help="If true, when an picking with type output having src_location contact and destination customers, a new picking with origin in the lot with quantity 1 is propesed to this location")

    @api.model
    def name_search(self, name, args=None, operator='ilike', limit=100):
        res = super(stock_location, self).name_search(name, args, operator, limit=20)
        return res


class stock_picking(models.Model):
    _inherit = 'stock.picking'

    default_dest_location_id = fields.Many2one('stock.location', string='Default Destination', domain=[('is_transit', '=', True)])

    @api.multi
    def apply_contact_picking(self, origin):
        """
        1. For each move, check where is the real stock.
            1.1 If in contact wh, skip as there is no necessity to do anything
            1.2 If in other internal location: Generate a new internal picking from that location to our transit location
                1.2.1 Force the assignation
        """
        self.ensure_one()
        # Get contact locations
        self._cr.execute("select pc.stock_location_id, pc.stock_location_transit_id from pos_order po join pos_session ps on ps.id=po.session_id join pos_config pc on pc.id=ps.config_id where po.picking_id=%(pick_id)s", {'pick_id': self.id})
        loc1, loc2 = self._cr.fetchone()
        location_ids = [loc1, loc2]
        for move in self.move_lines:
            pass
        todo_lot_lst = []
        # Collect the lots that needs to be transferred to our location.
        for op in self.pack_operation_ids:
            if op.location_dest_id.id == loc1:  # refund
                continue
            q_params = {
                'prodlot_id': op.lot_id.id,
                'location_ids': tuple(location_ids),
            }
            q = """
                SELECT COALESCE(SUM(qty), 0) FROM stock_quant
                WHERE lot_id=%(prodlot_id)s
                    AND location_id IN %(location_ids)s;
            """
            self._cr.execute(q, q_params)
            qty, = self._cr.fetchone()
            if qty < 0:
                todo_lot_lst.append(op.lot_id.id)
                logger.info("Generate a move for lot %s" % op.lot_id.name)
            else:
                logger.warning("Omit a move for lot %s, because the quantity of this lot in our location is non lower than zero" % op.lot_id.name)
        if todo_lot_lst:  # We group by location_id where the lot has enough qty
                picking_obj = self.env['stock.picking']
                move_obj = self.env['stock.move']
                prod_obj = self.env['product.product']
                quant_obj = self.env['stock.quant']
                q = """
                    SELECT sq.location_id, sq.lot_id, sq.id, sq.product_id
                    FROM stock_quant sq
                        JOIN stock_location sl ON sl.id=sq.location_id
                    WHERE sl.usage ~ 'internal' AND sq.lot_id = %(lot_ids)s AND sq.qty > 0
                """
                q_params = {
                    'lot_ids': tuple(todo_lot_lst),
                }
                self._cr.execute(q, q_params)
                lot_grouped_by_location = {}  # {location_id: [(lot_id, quant_id, product_id),], }
                for sq in self._cr.fetchall():
                    if sq[0] not in lot_grouped_by_location:
                        lot_grouped_by_location[sq[0]] = [(sq[1], sq[2], sq[3])]
                    else:
                        lot_grouped_by_location[sq[0]].append((sq[1], sq[2], sq[3]))
                # Make the pickings with source in location_id and destination transit of our wh
                for location, vals in lot_grouped_by_location.items():
                    source_location_id = location
                    dest_location_id = loc2  # Transit
                    # Get the picking type of internal with source as the origin of our movement
                    q = """
                        SELECT id FROM stock_picking_type
                        WHERE default_location_src_id=%(loc_id)s
                            AND active is true AND code ~ 'internal'
                    """
                    self._cr.execute(q, {'loc_id': location})
                    picking_type_id, = self._cr.fetchone()  # Internal Transfers of source_location_id
                    picking_id = picking_obj.create({
                        'origin': origin,
                        'picking_type_id': picking_type_id,
                    })
                    # Having the picking created, create the moves
                    for line in vals:
                        lot_id, quant_id, product_id = line
                        product = prod_obj.browse(product_id)
                        product_name = product.name_get()[0][0]

                        move_id = move_obj.create({
                            'name': product_name,
                            'product_uom': product.uom_id.id,
                            'product_uos': product.uom_id.id,
                            'picking_id': picking_id.id,
                            'picking_type_id': picking_type_id,
                            'product_id': product_id,
                            'product_uos_qty': 1,
                            'product_uom_qty': 1,
                            'state': 'draft',
                            'location_id': source_location_id,
                            'location_dest_id': dest_location_id,
                        })

                        # Reserve the quant for the move
                        quant = quant_obj.browse(quant_id)
                        quant_obj.quants_reserve([(quant, 1)], move_id)
                        # # Mark the move as assigned
                        # move_id.force_assign()

                    picking_id.action_confirm()
                    picking_id.force_assign()
                    # picking_obj.action_done([picking_id])
        return True

    @api.multi
    def action_done(self):
        res = super(stock_picking, self).do_transfer()
        for pick in self:
            q_params = {
                'picking_type_id': pick.picking_type_id.id,
                'customer_location': self.env.ref('stock.stock_location_customers').id,
            }
            q = """
            SELECT src_location.is_contact
            FROM stock_picking_type spt
                JOIN stock_location src_location ON src_location.id=spt.default_location_src_id
            WHERE spt.id=%(picking_type_id)s AND spt.default_location_dest_id=%(customer_location)s;
            """
            self._cr.execute(q, q_params)
            is_from_contact_to_customer, = self._cr.fetchone()
            if is_from_contact_to_customer:
                pick.sudo().apply_contact_picking(pick.name)  # Sudo to bypass ir.rules that gives problems
        return res

    @api.multi
    def action_confirm(self):
        res = super(stock_picking, self.sudo()).action_confirm()
        return res

    @api.cr_uid_ids_context
    def do_transfer(self, cr, uid, picking_ids, context=None):
        res = super(stock_picking, self).do_transfer(cr, uid, picking_ids, context)
        # Recompute locations of lots involved
        for picking in self.browse(cr, uid, picking_ids, context=context):
            lot_ids = picking.pack_operation_ids.mapped('lot_id')
            lot_ids.action_update_last_location()
        return res

    @api.model
    def fields_view_get(self, view_id=None, view_type='form', toolbar=False, submenu=False):
        res = super(stock_picking, self).fields_view_get(view_id=view_id, view_type=view_type, toolbar=toolbar, submenu=submenu)
        if view_type == 'form':
            spt_id = self._context.get('default_picking_type_id', False)
            if not spt_id:
                return res
            sp_id = self._context.get('active_ids')[0]

            # Obtenemos el spt_code
            if spt_id:
                spt_code = self.env['stock.picking.type'].browse(spt_id).code
            elif sp_id:
                spt_code = self.env['stock.picking'].browse(sp_id).picking_type_id.code

            if spt_code == 'internal':
                doc = etree.XML(res['arch'])
                for node in doc.xpath("//field[@name='default_dest_location_id']"):
                    node.set('invisible', '0')
                    node.set('modifiers', '{}')
                res['arch'] = etree.tostring(doc)
        return res

class stock_quant(models.Model):
    _inherit = 'stock.quant'

    @api.model
    def fields_view_get(self, view_id=None, view_type='form', toolbar=False, submenu=False):
        res = super(stock_quant, self).fields_view_get(view_id=view_id, view_type=view_type, toolbar=toolbar, submenu=submenu)
        # spt_id = self._context.get('default_picking_type_id')
        if view_type in ('search', 'tree') and self.env.user.pos_config:
            # spt = self.env['stock.picking.type'].browse(spt_id)
            poscfg = self.env.user.pos_config
            doc = etree.XML(res['arch'])
            locations_pos = [poscfg.stock_location_transit_id.id, poscfg.stock_location_id.id]
            for node in doc.xpath("//filter[@name='my_pos_location']"):
                domain = "[('location_id.id', 'in', %s)]" % locations_pos
                node.set('domain', domain)
            res['arch'] = etree.tostring(doc)
        return res

    @api.multi
    def write(self, values):
        res = super(stock_quant, self).write(values)
        if 'location_id' in values:
            self.lot_id.action_update_last_location()
        return res

    @api.model
    def create(self, values):
        res = super(stock_quant, self).create(values)
        if 'lot_id' in values and values['lot_id']:
            self.env['stock.production.lot'].browse(values['lot_id']).action_update_last_location()
        return res

    @api.multi
    def unlink(self):
        mylot = self.lot_id
        res = super(stock_quant, self).unlink()
        mylot.action_update_last_location()
        return res
